# Generated from _qpd_antlr/sql.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .sqlParser import sqlParser
else:
    from sqlParser import sqlParser

# This class defines a complete generic visitor for a parse tree produced by sqlParser.

class sqlVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by sqlParser#singleStatement.
    def visitSingleStatement(self, ctx:sqlParser.SingleStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleExpression.
    def visitSingleExpression(self, ctx:sqlParser.SingleExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleTableIdentifier.
    def visitSingleTableIdentifier(self, ctx:sqlParser.SingleTableIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleMultipartIdentifier.
    def visitSingleMultipartIdentifier(self, ctx:sqlParser.SingleMultipartIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleFunctionIdentifier.
    def visitSingleFunctionIdentifier(self, ctx:sqlParser.SingleFunctionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleDataType.
    def visitSingleDataType(self, ctx:sqlParser.SingleDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleTableSchema.
    def visitSingleTableSchema(self, ctx:sqlParser.SingleTableSchemaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#statementDefault.
    def visitStatementDefault(self, ctx:sqlParser.StatementDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dmlStatement.
    def visitDmlStatement(self, ctx:sqlParser.DmlStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#use.
    def visitUse(self, ctx:sqlParser.UseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createNamespace.
    def visitCreateNamespace(self, ctx:sqlParser.CreateNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setNamespaceProperties.
    def visitSetNamespaceProperties(self, ctx:sqlParser.SetNamespacePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setNamespaceLocation.
    def visitSetNamespaceLocation(self, ctx:sqlParser.SetNamespaceLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropNamespace.
    def visitDropNamespace(self, ctx:sqlParser.DropNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showNamespaces.
    def visitShowNamespaces(self, ctx:sqlParser.ShowNamespacesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createTable.
    def visitCreateTable(self, ctx:sqlParser.CreateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createHiveTable.
    def visitCreateHiveTable(self, ctx:sqlParser.CreateHiveTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createTableLike.
    def visitCreateTableLike(self, ctx:sqlParser.CreateTableLikeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#replaceTable.
    def visitReplaceTable(self, ctx:sqlParser.ReplaceTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#analyze.
    def visitAnalyze(self, ctx:sqlParser.AnalyzeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#addTableColumns.
    def visitAddTableColumns(self, ctx:sqlParser.AddTableColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#renameTableColumn.
    def visitRenameTableColumn(self, ctx:sqlParser.RenameTableColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropTableColumns.
    def visitDropTableColumns(self, ctx:sqlParser.DropTableColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#renameTable.
    def visitRenameTable(self, ctx:sqlParser.RenameTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setTableProperties.
    def visitSetTableProperties(self, ctx:sqlParser.SetTablePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#unsetTableProperties.
    def visitUnsetTableProperties(self, ctx:sqlParser.UnsetTablePropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#alterTableAlterColumn.
    def visitAlterTableAlterColumn(self, ctx:sqlParser.AlterTableAlterColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#hiveChangeColumn.
    def visitHiveChangeColumn(self, ctx:sqlParser.HiveChangeColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#hiveReplaceColumns.
    def visitHiveReplaceColumns(self, ctx:sqlParser.HiveReplaceColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setTableSerDe.
    def visitSetTableSerDe(self, ctx:sqlParser.SetTableSerDeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#addTablePartition.
    def visitAddTablePartition(self, ctx:sqlParser.AddTablePartitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#renameTablePartition.
    def visitRenameTablePartition(self, ctx:sqlParser.RenameTablePartitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropTablePartitions.
    def visitDropTablePartitions(self, ctx:sqlParser.DropTablePartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setTableLocation.
    def visitSetTableLocation(self, ctx:sqlParser.SetTableLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#recoverPartitions.
    def visitRecoverPartitions(self, ctx:sqlParser.RecoverPartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropTable.
    def visitDropTable(self, ctx:sqlParser.DropTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropView.
    def visitDropView(self, ctx:sqlParser.DropViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createView.
    def visitCreateView(self, ctx:sqlParser.CreateViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createTempViewUsing.
    def visitCreateTempViewUsing(self, ctx:sqlParser.CreateTempViewUsingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#alterViewQuery.
    def visitAlterViewQuery(self, ctx:sqlParser.AlterViewQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createFunction.
    def visitCreateFunction(self, ctx:sqlParser.CreateFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dropFunction.
    def visitDropFunction(self, ctx:sqlParser.DropFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#explain.
    def visitExplain(self, ctx:sqlParser.ExplainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showTables.
    def visitShowTables(self, ctx:sqlParser.ShowTablesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showTable.
    def visitShowTable(self, ctx:sqlParser.ShowTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showTblProperties.
    def visitShowTblProperties(self, ctx:sqlParser.ShowTblPropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showColumns.
    def visitShowColumns(self, ctx:sqlParser.ShowColumnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showViews.
    def visitShowViews(self, ctx:sqlParser.ShowViewsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showPartitions.
    def visitShowPartitions(self, ctx:sqlParser.ShowPartitionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showFunctions.
    def visitShowFunctions(self, ctx:sqlParser.ShowFunctionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showCreateTable.
    def visitShowCreateTable(self, ctx:sqlParser.ShowCreateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#showCurrentNamespace.
    def visitShowCurrentNamespace(self, ctx:sqlParser.ShowCurrentNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeFunction.
    def visitDescribeFunction(self, ctx:sqlParser.DescribeFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeNamespace.
    def visitDescribeNamespace(self, ctx:sqlParser.DescribeNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeRelation.
    def visitDescribeRelation(self, ctx:sqlParser.DescribeRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeQuery.
    def visitDescribeQuery(self, ctx:sqlParser.DescribeQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#commentNamespace.
    def visitCommentNamespace(self, ctx:sqlParser.CommentNamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#commentTable.
    def visitCommentTable(self, ctx:sqlParser.CommentTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#refreshTable.
    def visitRefreshTable(self, ctx:sqlParser.RefreshTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#refreshResource.
    def visitRefreshResource(self, ctx:sqlParser.RefreshResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#cacheTable.
    def visitCacheTable(self, ctx:sqlParser.CacheTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#uncacheTable.
    def visitUncacheTable(self, ctx:sqlParser.UncacheTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#clearCache.
    def visitClearCache(self, ctx:sqlParser.ClearCacheContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#loadData.
    def visitLoadData(self, ctx:sqlParser.LoadDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#truncateTable.
    def visitTruncateTable(self, ctx:sqlParser.TruncateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#repairTable.
    def visitRepairTable(self, ctx:sqlParser.RepairTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#manageResource.
    def visitManageResource(self, ctx:sqlParser.ManageResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#failNativeCommand.
    def visitFailNativeCommand(self, ctx:sqlParser.FailNativeCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setConfiguration.
    def visitSetConfiguration(self, ctx:sqlParser.SetConfigurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#resetConfiguration.
    def visitResetConfiguration(self, ctx:sqlParser.ResetConfigurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#unsupportedHiveNativeCommands.
    def visitUnsupportedHiveNativeCommands(self, ctx:sqlParser.UnsupportedHiveNativeCommandsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createTableHeader.
    def visitCreateTableHeader(self, ctx:sqlParser.CreateTableHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#replaceTableHeader.
    def visitReplaceTableHeader(self, ctx:sqlParser.ReplaceTableHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#bucketSpec.
    def visitBucketSpec(self, ctx:sqlParser.BucketSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#skewSpec.
    def visitSkewSpec(self, ctx:sqlParser.SkewSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#locationSpec.
    def visitLocationSpec(self, ctx:sqlParser.LocationSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#commentSpec.
    def visitCommentSpec(self, ctx:sqlParser.CommentSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#query.
    def visitQuery(self, ctx:sqlParser.QueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#insertOverwriteTable.
    def visitInsertOverwriteTable(self, ctx:sqlParser.InsertOverwriteTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#insertIntoTable.
    def visitInsertIntoTable(self, ctx:sqlParser.InsertIntoTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#insertOverwriteHiveDir.
    def visitInsertOverwriteHiveDir(self, ctx:sqlParser.InsertOverwriteHiveDirContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#insertOverwriteDir.
    def visitInsertOverwriteDir(self, ctx:sqlParser.InsertOverwriteDirContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#partitionSpecLocation.
    def visitPartitionSpecLocation(self, ctx:sqlParser.PartitionSpecLocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#partitionSpec.
    def visitPartitionSpec(self, ctx:sqlParser.PartitionSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#partitionVal.
    def visitPartitionVal(self, ctx:sqlParser.PartitionValContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#namespace.
    def visitNamespace(self, ctx:sqlParser.NamespaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeFuncName.
    def visitDescribeFuncName(self, ctx:sqlParser.DescribeFuncNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#describeColName.
    def visitDescribeColName(self, ctx:sqlParser.DescribeColNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#ctes.
    def visitCtes(self, ctx:sqlParser.CtesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#namedQuery.
    def visitNamedQuery(self, ctx:sqlParser.NamedQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableProvider.
    def visitTableProvider(self, ctx:sqlParser.TableProviderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createTableClauses.
    def visitCreateTableClauses(self, ctx:sqlParser.CreateTableClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tablePropertyList.
    def visitTablePropertyList(self, ctx:sqlParser.TablePropertyListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableProperty.
    def visitTableProperty(self, ctx:sqlParser.TablePropertyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tablePropertyKey.
    def visitTablePropertyKey(self, ctx:sqlParser.TablePropertyKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tablePropertyValue.
    def visitTablePropertyValue(self, ctx:sqlParser.TablePropertyValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#constantList.
    def visitConstantList(self, ctx:sqlParser.ConstantListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#nestedConstantList.
    def visitNestedConstantList(self, ctx:sqlParser.NestedConstantListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#createFileFormat.
    def visitCreateFileFormat(self, ctx:sqlParser.CreateFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableFileFormat.
    def visitTableFileFormat(self, ctx:sqlParser.TableFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#genericFileFormat.
    def visitGenericFileFormat(self, ctx:sqlParser.GenericFileFormatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#storageHandler.
    def visitStorageHandler(self, ctx:sqlParser.StorageHandlerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#resource.
    def visitResource(self, ctx:sqlParser.ResourceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#singleInsertQuery.
    def visitSingleInsertQuery(self, ctx:sqlParser.SingleInsertQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#multiInsertQuery.
    def visitMultiInsertQuery(self, ctx:sqlParser.MultiInsertQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#deleteFromTable.
    def visitDeleteFromTable(self, ctx:sqlParser.DeleteFromTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#updateTable.
    def visitUpdateTable(self, ctx:sqlParser.UpdateTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#mergeIntoTable.
    def visitMergeIntoTable(self, ctx:sqlParser.MergeIntoTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#queryOrganization.
    def visitQueryOrganization(self, ctx:sqlParser.QueryOrganizationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#multiInsertQueryBody.
    def visitMultiInsertQueryBody(self, ctx:sqlParser.MultiInsertQueryBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#queryTermDefault.
    def visitQueryTermDefault(self, ctx:sqlParser.QueryTermDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setOperation.
    def visitSetOperation(self, ctx:sqlParser.SetOperationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#queryPrimaryDefault.
    def visitQueryPrimaryDefault(self, ctx:sqlParser.QueryPrimaryDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#fromStmt.
    def visitFromStmt(self, ctx:sqlParser.FromStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#table.
    def visitTable(self, ctx:sqlParser.TableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#inlineTableDefault1.
    def visitInlineTableDefault1(self, ctx:sqlParser.InlineTableDefault1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#subquery.
    def visitSubquery(self, ctx:sqlParser.SubqueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sortItem.
    def visitSortItem(self, ctx:sqlParser.SortItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#fromStatement.
    def visitFromStatement(self, ctx:sqlParser.FromStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#fromStatementBody.
    def visitFromStatementBody(self, ctx:sqlParser.FromStatementBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#transformQuerySpecification.
    def visitTransformQuerySpecification(self, ctx:sqlParser.TransformQuerySpecificationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#regularQuerySpecification.
    def visitRegularQuerySpecification(self, ctx:sqlParser.RegularQuerySpecificationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#transformClause.
    def visitTransformClause(self, ctx:sqlParser.TransformClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#selectClause.
    def visitSelectClause(self, ctx:sqlParser.SelectClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setClause.
    def visitSetClause(self, ctx:sqlParser.SetClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#matchedClause.
    def visitMatchedClause(self, ctx:sqlParser.MatchedClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#notMatchedClause.
    def visitNotMatchedClause(self, ctx:sqlParser.NotMatchedClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#matchedAction.
    def visitMatchedAction(self, ctx:sqlParser.MatchedActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#notMatchedAction.
    def visitNotMatchedAction(self, ctx:sqlParser.NotMatchedActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#assignmentList.
    def visitAssignmentList(self, ctx:sqlParser.AssignmentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#assignment.
    def visitAssignment(self, ctx:sqlParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#whereClause.
    def visitWhereClause(self, ctx:sqlParser.WhereClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#havingClause.
    def visitHavingClause(self, ctx:sqlParser.HavingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#hint.
    def visitHint(self, ctx:sqlParser.HintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#hintStatement.
    def visitHintStatement(self, ctx:sqlParser.HintStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#fromClause.
    def visitFromClause(self, ctx:sqlParser.FromClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#aggregationClause.
    def visitAggregationClause(self, ctx:sqlParser.AggregationClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#groupingSet.
    def visitGroupingSet(self, ctx:sqlParser.GroupingSetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#pivotClause.
    def visitPivotClause(self, ctx:sqlParser.PivotClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#pivotColumn.
    def visitPivotColumn(self, ctx:sqlParser.PivotColumnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#pivotValue.
    def visitPivotValue(self, ctx:sqlParser.PivotValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#lateralView.
    def visitLateralView(self, ctx:sqlParser.LateralViewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#setQuantifier.
    def visitSetQuantifier(self, ctx:sqlParser.SetQuantifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#relation.
    def visitRelation(self, ctx:sqlParser.RelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#joinRelation.
    def visitJoinRelation(self, ctx:sqlParser.JoinRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#joinType.
    def visitJoinType(self, ctx:sqlParser.JoinTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#joinCriteria.
    def visitJoinCriteria(self, ctx:sqlParser.JoinCriteriaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sample.
    def visitSample(self, ctx:sqlParser.SampleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sampleByPercentile.
    def visitSampleByPercentile(self, ctx:sqlParser.SampleByPercentileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sampleByRows.
    def visitSampleByRows(self, ctx:sqlParser.SampleByRowsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sampleByBucket.
    def visitSampleByBucket(self, ctx:sqlParser.SampleByBucketContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#sampleByBytes.
    def visitSampleByBytes(self, ctx:sqlParser.SampleByBytesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identifierList.
    def visitIdentifierList(self, ctx:sqlParser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identifierSeq.
    def visitIdentifierSeq(self, ctx:sqlParser.IdentifierSeqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#orderedIdentifierList.
    def visitOrderedIdentifierList(self, ctx:sqlParser.OrderedIdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#orderedIdentifier.
    def visitOrderedIdentifier(self, ctx:sqlParser.OrderedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identifierCommentList.
    def visitIdentifierCommentList(self, ctx:sqlParser.IdentifierCommentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identifierComment.
    def visitIdentifierComment(self, ctx:sqlParser.IdentifierCommentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableName.
    def visitTableName(self, ctx:sqlParser.TableNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#aliasedQuery.
    def visitAliasedQuery(self, ctx:sqlParser.AliasedQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#aliasedRelation.
    def visitAliasedRelation(self, ctx:sqlParser.AliasedRelationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#inlineTable.
    def visitInlineTable(self, ctx:sqlParser.InlineTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#functionTable.
    def visitFunctionTable(self, ctx:sqlParser.FunctionTableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableAlias.
    def visitTableAlias(self, ctx:sqlParser.TableAliasContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#rowFormatSerde.
    def visitRowFormatSerde(self, ctx:sqlParser.RowFormatSerdeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#rowFormatDelimited.
    def visitRowFormatDelimited(self, ctx:sqlParser.RowFormatDelimitedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#multipartIdentifierList.
    def visitMultipartIdentifierList(self, ctx:sqlParser.MultipartIdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#multipartIdentifier.
    def visitMultipartIdentifier(self, ctx:sqlParser.MultipartIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tableIdentifier.
    def visitTableIdentifier(self, ctx:sqlParser.TableIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#functionIdentifier.
    def visitFunctionIdentifier(self, ctx:sqlParser.FunctionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#namedExpression.
    def visitNamedExpression(self, ctx:sqlParser.NamedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#namedExpressionSeq.
    def visitNamedExpressionSeq(self, ctx:sqlParser.NamedExpressionSeqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#transformList.
    def visitTransformList(self, ctx:sqlParser.TransformListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identityTransform.
    def visitIdentityTransform(self, ctx:sqlParser.IdentityTransformContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#applyTransform.
    def visitApplyTransform(self, ctx:sqlParser.ApplyTransformContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#transformArgument.
    def visitTransformArgument(self, ctx:sqlParser.TransformArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#expression.
    def visitExpression(self, ctx:sqlParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#logicalNot.
    def visitLogicalNot(self, ctx:sqlParser.LogicalNotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#predicated.
    def visitPredicated(self, ctx:sqlParser.PredicatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#exists.
    def visitExists(self, ctx:sqlParser.ExistsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#logicalBinary.
    def visitLogicalBinary(self, ctx:sqlParser.LogicalBinaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#predicate.
    def visitPredicate(self, ctx:sqlParser.PredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#valueExpressionDefault.
    def visitValueExpressionDefault(self, ctx:sqlParser.ValueExpressionDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#comparison.
    def visitComparison(self, ctx:sqlParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#arithmeticBinary.
    def visitArithmeticBinary(self, ctx:sqlParser.ArithmeticBinaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#arithmeticUnary.
    def visitArithmeticUnary(self, ctx:sqlParser.ArithmeticUnaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#struct.
    def visitStruct(self, ctx:sqlParser.StructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#dereference.
    def visitDereference(self, ctx:sqlParser.DereferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#simpleCase.
    def visitSimpleCase(self, ctx:sqlParser.SimpleCaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#columnReference.
    def visitColumnReference(self, ctx:sqlParser.ColumnReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#rowConstructor.
    def visitRowConstructor(self, ctx:sqlParser.RowConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#last.
    def visitLast(self, ctx:sqlParser.LastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#star.
    def visitStar(self, ctx:sqlParser.StarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#overlay.
    def visitOverlay(self, ctx:sqlParser.OverlayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#subscript.
    def visitSubscript(self, ctx:sqlParser.SubscriptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#subqueryExpression.
    def visitSubqueryExpression(self, ctx:sqlParser.SubqueryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#substring.
    def visitSubstring(self, ctx:sqlParser.SubstringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#currentDatetime.
    def visitCurrentDatetime(self, ctx:sqlParser.CurrentDatetimeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#cast.
    def visitCast(self, ctx:sqlParser.CastContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#constantDefault.
    def visitConstantDefault(self, ctx:sqlParser.ConstantDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#lambda.
    def visitLambda(self, ctx:sqlParser.LambdaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#parenthesizedExpression.
    def visitParenthesizedExpression(self, ctx:sqlParser.ParenthesizedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#extract.
    def visitExtract(self, ctx:sqlParser.ExtractContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#trim.
    def visitTrim(self, ctx:sqlParser.TrimContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#functionCall.
    def visitFunctionCall(self, ctx:sqlParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#searchedCase.
    def visitSearchedCase(self, ctx:sqlParser.SearchedCaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#position.
    def visitPosition(self, ctx:sqlParser.PositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#first.
    def visitFirst(self, ctx:sqlParser.FirstContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#nullLiteral.
    def visitNullLiteral(self, ctx:sqlParser.NullLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#intervalLiteral.
    def visitIntervalLiteral(self, ctx:sqlParser.IntervalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#typeConstructor.
    def visitTypeConstructor(self, ctx:sqlParser.TypeConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#numericLiteral.
    def visitNumericLiteral(self, ctx:sqlParser.NumericLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#booleanLiteral.
    def visitBooleanLiteral(self, ctx:sqlParser.BooleanLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#stringLiteral.
    def visitStringLiteral(self, ctx:sqlParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#comparisonOperator.
    def visitComparisonOperator(self, ctx:sqlParser.ComparisonOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#arithmeticOperator.
    def visitArithmeticOperator(self, ctx:sqlParser.ArithmeticOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#predicateOperator.
    def visitPredicateOperator(self, ctx:sqlParser.PredicateOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#booleanValue.
    def visitBooleanValue(self, ctx:sqlParser.BooleanValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#interval.
    def visitInterval(self, ctx:sqlParser.IntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#errorCapturingMultiUnitsInterval.
    def visitErrorCapturingMultiUnitsInterval(self, ctx:sqlParser.ErrorCapturingMultiUnitsIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#multiUnitsInterval.
    def visitMultiUnitsInterval(self, ctx:sqlParser.MultiUnitsIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#errorCapturingUnitToUnitInterval.
    def visitErrorCapturingUnitToUnitInterval(self, ctx:sqlParser.ErrorCapturingUnitToUnitIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#unitToUnitInterval.
    def visitUnitToUnitInterval(self, ctx:sqlParser.UnitToUnitIntervalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#intervalValue.
    def visitIntervalValue(self, ctx:sqlParser.IntervalValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#intervalUnit.
    def visitIntervalUnit(self, ctx:sqlParser.IntervalUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#colPosition.
    def visitColPosition(self, ctx:sqlParser.ColPositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#complexDataType.
    def visitComplexDataType(self, ctx:sqlParser.ComplexDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#primitiveDataType.
    def visitPrimitiveDataType(self, ctx:sqlParser.PrimitiveDataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#qualifiedColTypeWithPositionList.
    def visitQualifiedColTypeWithPositionList(self, ctx:sqlParser.QualifiedColTypeWithPositionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#qualifiedColTypeWithPosition.
    def visitQualifiedColTypeWithPosition(self, ctx:sqlParser.QualifiedColTypeWithPositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#colTypeList.
    def visitColTypeList(self, ctx:sqlParser.ColTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#colType.
    def visitColType(self, ctx:sqlParser.ColTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#complexColTypeList.
    def visitComplexColTypeList(self, ctx:sqlParser.ComplexColTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#complexColType.
    def visitComplexColType(self, ctx:sqlParser.ComplexColTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#whenClause.
    def visitWhenClause(self, ctx:sqlParser.WhenClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#windowClause.
    def visitWindowClause(self, ctx:sqlParser.WindowClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#namedWindow.
    def visitNamedWindow(self, ctx:sqlParser.NamedWindowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#windowRef.
    def visitWindowRef(self, ctx:sqlParser.WindowRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#windowDef.
    def visitWindowDef(self, ctx:sqlParser.WindowDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#windowFrame.
    def visitWindowFrame(self, ctx:sqlParser.WindowFrameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#frameBound.
    def visitFrameBound(self, ctx:sqlParser.FrameBoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#qualifiedNameList.
    def visitQualifiedNameList(self, ctx:sqlParser.QualifiedNameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#functionName.
    def visitFunctionName(self, ctx:sqlParser.FunctionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#qualifiedName.
    def visitQualifiedName(self, ctx:sqlParser.QualifiedNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#errorCapturingIdentifier.
    def visitErrorCapturingIdentifier(self, ctx:sqlParser.ErrorCapturingIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#errorIdent.
    def visitErrorIdent(self, ctx:sqlParser.ErrorIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#realIdent.
    def visitRealIdent(self, ctx:sqlParser.RealIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#identifier.
    def visitIdentifier(self, ctx:sqlParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#unquotedIdentifier.
    def visitUnquotedIdentifier(self, ctx:sqlParser.UnquotedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#quotedIdentifierAlternative.
    def visitQuotedIdentifierAlternative(self, ctx:sqlParser.QuotedIdentifierAlternativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#quotedIdentifier.
    def visitQuotedIdentifier(self, ctx:sqlParser.QuotedIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#exponentLiteral.
    def visitExponentLiteral(self, ctx:sqlParser.ExponentLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#decimalLiteral.
    def visitDecimalLiteral(self, ctx:sqlParser.DecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#legacyDecimalLiteral.
    def visitLegacyDecimalLiteral(self, ctx:sqlParser.LegacyDecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#integerLiteral.
    def visitIntegerLiteral(self, ctx:sqlParser.IntegerLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#bigIntLiteral.
    def visitBigIntLiteral(self, ctx:sqlParser.BigIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#smallIntLiteral.
    def visitSmallIntLiteral(self, ctx:sqlParser.SmallIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#tinyIntLiteral.
    def visitTinyIntLiteral(self, ctx:sqlParser.TinyIntLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#doubleLiteral.
    def visitDoubleLiteral(self, ctx:sqlParser.DoubleLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#bigDecimalLiteral.
    def visitBigDecimalLiteral(self, ctx:sqlParser.BigDecimalLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#alterColumnAction.
    def visitAlterColumnAction(self, ctx:sqlParser.AlterColumnActionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#ansiNonReserved.
    def visitAnsiNonReserved(self, ctx:sqlParser.AnsiNonReservedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#strictNonReserved.
    def visitStrictNonReserved(self, ctx:sqlParser.StrictNonReservedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by sqlParser#nonReserved.
    def visitNonReserved(self, ctx:sqlParser.NonReservedContext):
        return self.visitChildren(ctx)



del sqlParser


from qpd_test import assert_df_eq
import pandas as pd
from pytest import raises


def test_assert_df_eq():
    a = pd.DataFrame([[0, 1]], columns=["a", "b"])
    assert_df_eq(a, a)
    assert_df_eq(a, [[0.0, 1.0]], ["a", "b"])
    assert not assert_df_eq(a, [[0.001, 1.001]], ["a", "b"], throw=False)
    assert_df_eq(a, [[0.001, 1.001]], ["a", "b"], digits=1)
    assert_df_eq(a, [[1, 0]], ["b", "a"])
    assert not assert_df_eq(a, [[1, 0]], ["b", "a"], check_col_order=True, throw=False)
    assert not assert_df_eq(a, [[10, 0]], ["b", "a"], throw=False)
    assert_df_eq(a, [[10, 0]], ["b", "a"], check_content=False)
    with raises(AssertionError):
        assert_df_eq(a, [[1.0, 1.0]], ["a", "b"])

    a = pd.DataFrame([[0, 1], [3, 4]], columns=["a", "b"])
    b = pd.DataFrame([[3, 4], [0, 1]], columns=["a", "b"])
    assert_df_eq(a, b)
    assert not assert_df_eq(a, b, check_order=True, throw=False)



from typing import Any, Callable, Dict, List, Tuple

import dask.array as np
import dask.dataframe as pd
import numpy
import pandas
from qpd import PandasLikeUtils, QPDEngine, run_sql
from qpd.dataframe import Column, DataFrame
from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    OrderBySpec,
    RollingWindowFrame,
    WindowFunctionSpec,
)
from qpd_pandas.engine import _RowsIndexer
from triad.utils.assertion import assert_or_throw


def run_sql_on_dask(
    sql: str, dfs: Dict[str, Any], ignore_case: bool = False
) -> pd.DataFrame:
    return run_sql(QPDDaskEngine(), sql, dfs, ignore_case=ignore_case)


class DaskUtils(PandasLikeUtils[pd.DataFrame, pd.Series]):
    pass


class QPDDaskEngine(QPDEngine):
    @property
    def pl_utils(self) -> DaskUtils:
        return DaskUtils()

    def to_df(self, obj: Any) -> DataFrame:
        if isinstance(obj, DataFrame):
            return obj
        if isinstance(obj, pandas.DataFrame):
            df = pd.from_pandas(obj, npartitions=2)  # TODO: don't hard code
            return DataFrame(Column(df[x], x) for x in df.columns.tolist())
        if isinstance(obj, pd.DataFrame):
            return DataFrame(Column(obj[x], x) for x in obj.columns.tolist())
        raise ValueError(f"{obj} is not supported")  # pragma: no cover

    def to_col(self, value: Any, name: str = "") -> Column:
        if isinstance(value, list):
            pdf = pandas.DataFrame({name: value})
            ddf = pd.from_pandas(pdf, npartitions=2)
            return Column(ddf[name], name)
        return Column(value, name)

    def to_native(self, df: DataFrame) -> Any:
        if all(not self.is_series(x.native) for x in df.values()):
            return pd.from_pandas(
                pandas.DataFrame({k: [v.native] for k, v in df.items()}), npartitions=2
            )
        values = {k: v.native for k, v in df.items() if not self.is_series(v.native)}
        series = [v.native.rename(k) for k, v in df.items() if self.is_series(v.native)]
        assert_or_throw(
            len(series) > 0, ValueError("DataFrame doesn't contain any series")
        )
        ddf = series[0].to_frame()
        for s in series[1:]:
            ddf[s.name] = s
        for k, v in values.items():
            ddf[k] = v
        return ddf[list(df.keys())]

    def is_series(self, obj: Any) -> bool:
        return isinstance(obj, pd.Series)

    def case_when(self, *cols: Column) -> Column:
        """`cols` must be in the format of
        `when1`, `value1`, ... ,`default`, and length must be >= 3.
        The reason to design the interface in this way is to simplify the translation
        from SQL to engine APIs.

        :return: [description]
        :rtype: Column
        """
        assert_or_throw(len(cols) >= 3, "case when must have 1+ conditions")

        # construct a new dataframe with all conditions and values
        # we have to do this to keep all series/consts with the same dimension
        # dask array where will lose shape info
        # so we have to make this operation on dataframe
        series: List[Tuple[int, pd.Series]] = []
        values: List[Tuple[int, Any]] = []
        for i in range(len(cols)):
            if self.is_series(cols[i].native):
                series.append((i, cols[i].native))
            else:
                values.append((i, cols[i].native))
        assert_or_throw(
            len(series) > 0, ValueError("DataFrame doesn't contain any series")
        )
        ddf = series[0][1].to_frame(name=str(series[0][0]))
        for i, s in series[1:]:
            ddf[str(i)] = s
        for i, v in values:
            ddf[str(i)] = v

        # adding masks backwards
        pos = len(cols) - 1
        result = ddf[str(pos)]
        pos -= 2
        while pos >= 0:
            result = result.mask(self._safe_bool(ddf[str(pos)]) > 0, ddf[str(pos + 1)])
            pos -= 2
        return Column(result)

    def order_by_limit(
        self, df: DataFrame, order_by: OrderBySpec, limit: int
    ) -> DataFrame:
        assert_or_throw(
            not (len(order_by) > 0 and limit < 0),
            ValueError("for dask engine, limit is required by order by"),
        )
        if len(order_by) == 0 and limit < 0:  # pragma: no cover
            return df
        ndf = self.to_native(df)
        if len(order_by) == 0 or limit == 0:
            return self.to_df(ndf.head(limit, npartitions=-1, compute=False))

        ndf, sort_keys, asc = self._prepare_sort(ndf, order_by)

        def p_apply(df: Any) -> Any:
            return df.sort_values(sort_keys, ascending=asc).head(limit)

        meta = [(ndf[x].name, ndf[x].dtype) for x in ndf.columns]
        res = ndf.map_partitions(p_apply, meta=meta).compute()
        res = res.sort_values(sort_keys, ascending=asc).head(limit)
        return self.to_df(res[list(df.keys())])

    def group_agg(
        self,
        df: DataFrame,
        keys: List[str],
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        if len(keys) > 0:
            return self._group_agg_with_keys(df, keys, agg_map)
        else:
            return self._group_agg_no_keys(df, agg_map)

    def window(  # noqa: C901
        self,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        if func.name == "row_number":
            return self._window_row_number(df, func, args, dest_col_name)
        if func.name == "sum":
            return self._window_simple_agg(
                lambda x: x.sum(min_count=1), df, func, args, dest_col_name
            )
        if func.name == "min":
            return self._window_simple_agg(
                lambda x: x.min(), df, func, args, dest_col_name
            )
        if func.name == "max":
            return self._window_simple_agg(
                lambda x: x.max(), df, func, args, dest_col_name
            )
        if func.name == "count":
            return self._window_simple_agg(
                lambda x: x.count(), df, func, args, dest_col_name
            )
        if func.name in ["avg", "mean"]:
            return self._window_simple_agg(
                lambda x: x.mean(), df, func, args, dest_col_name
            )
        if func.name in ["rank"]:
            return self._window_ranks(
                lambda x: x.rank(method="min"), df, func, args, dest_col_name
            )
        if func.name in ["dense_rank"]:
            return self._window_ranks(
                lambda x: x.rank(method="dense"), df, func, args, dest_col_name
            )
        if func.name in ["percentile_rank", "percent_rank"]:

            def pct_rank(s: Any) -> Any:
                m = s.size - 1
                ranks = s.rank(method="min") - 1
                if m == 0:
                    m = 1.0
                return ranks / m

            return self._window_ranks(
                lambda x: pct_rank(x)
                if isinstance(x, pandas.Series)
                else x.apply(pct_rank),
                df,
                func,
                args,
                dest_col_name,
            )
        if func.name == "lead":
            return self._window_lead_lag(
                lambda x, delta, default: x.shift(-delta, fill_value=default),
                df,
                func,
                args,
                dest_col_name,
            )
        if func.name == "lag":
            return self._window_lead_lag(
                lambda x, delta, default: x.shift(delta, fill_value=default),
                df,
                func,
                args,
                dest_col_name,
            )
        raise NotImplementedError(f"{func}")  # pragma: no cover

    def _window_row_number(
        self,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) == 0, ValueError(f"{args}"))
        assert_or_throw(
            not func.has_windowframe, ValueError("row_number can't have windowframe")
        )
        assert_or_throw(
            not (func.has_order_by and not func.has_partition_by),
            ValueError("for dask engine, order by requires partition by"),
        )
        assert_or_throw(
            func.has_partition_by,
            ValueError("for dask engine, partition by is required for row_number()"),
        )
        keys = list(df.keys())
        ndf = self.to_native(df)
        ndf, sort_keys, asc = self._prepare_sort(ndf, func.window.order_by)
        gp, gp_keys, _ = self._safe_groupby(ndf, func.window.partition_keys)
        if not func.has_order_by:  # pragma: no cover
            # unable to have deterministic test
            col = gp.cumcount() + 1
            ndf[dest_col_name] = col
            return self.to_df(ndf[keys + [dest_col_name]])

        def gp_apply(df: Any) -> Any:
            df = df.sort_values(sort_keys, ascending=asc)
            col = np.arange(1, 1 + df.shape[0])
            df[dest_col_name] = col
            return df.reset_index(drop=True)

        meta = [(ndf[x].name, ndf[x].dtype) for x in ndf.columns]
        meta += [(dest_col_name, "int64")]

        ndf = gp.apply(gp_apply, meta=meta).reset_index(drop=True)
        return self.to_df(ndf[keys + [dest_col_name]])

    def _window_ranks(
        self,
        rank_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) == 0, ValueError(f"{args}"))
        assert_or_throw(
            func.has_order_by, ValueError(f"rank functions require order by {func}")
        )
        assert_or_throw(
            not func.has_windowframe,
            ValueError(f"rank functions can't have windowframe {func}"),
        )
        assert_or_throw(
            not (func.has_order_by and not func.has_partition_by),
            ValueError("for dask engine, order by requires partition by"),
        )
        assert_or_throw(
            func.has_partition_by,
            ValueError("for dask engine, partition by is required for ranking"),
        )
        ndf = self.to_native(df)
        keys = list(df.keys())
        ndf, sort_keys, asc = self._prepare_sort(ndf, func.window.order_by)
        gp, gp_keys, _ = self._safe_groupby(ndf, func.window.partition_keys)

        def gp_apply(gdf: Any) -> Any:
            gdf = gdf.sort_values(sort_keys, ascending=asc)
            rank_on = self._safe_pandas_groupby(
                gdf, sort_keys, as_index=False, sort=False
            )[0].ngroup()
            gdf[dest_col_name] = rank_func(rank_on)
            return gdf.reset_index(drop=True)

        meta = [(ndf[x].name, ndf[x].dtype) for x in ndf.columns]
        meta += [(dest_col_name, "f8")]

        ndf = gp.apply(gp_apply, meta=meta).reset_index(drop=True)
        return self.to_df(ndf[keys + [dest_col_name]])

    def _window_lead_lag(
        self,
        l_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(
            not (func.has_order_by and not func.has_partition_by),
            ValueError("for dask engine, order by requires partition by"),
        )
        assert_or_throw(
            func.has_partition_by,
            ValueError("for dask engine, partition by is required for lead/lag"),
        )
        assert_or_throw(len(args) in [2, 3], ValueError(f"{args}"))
        assert_or_throw(args[0].is_col, ValueError(f"{args[0]}"))
        assert_or_throw(
            not args[1].is_col and isinstance(args[1].value, int),
            ValueError(f"{args[1]}"),
        )
        if len(args) == 3:
            assert_or_throw(not args[2].is_col, ValueError(f"{args[2]}"))
        assert_or_throw(
            not func.has_windowframe,
            ValueError(f"lead/lag functions can't have windowframe {func}"),
        )
        col = args[0].value
        delta = int(args[1].value)
        default = None if len(args) == 2 else args[2].value

        ndf = self.to_native(df)
        keys = list(df.keys())
        ndf, sort_keys, asc = self._prepare_sort(ndf, func.window.order_by)
        gp, gp_keys, _ = self._safe_groupby(ndf, func.window.partition_keys)

        def gp_apply(gdf: Any) -> Any:
            gdf = gdf.sort_values(sort_keys, ascending=asc)
            gdf[dest_col_name] = l_func(gdf[col], delta, default)
            return gdf.reset_index(drop=True)

        meta = [(ndf[x].name, ndf[x].dtype) for x in ndf.columns]
        meta += [(dest_col_name, ndf[col].dtype)]

        ndf = gp.apply(gp_apply, meta=meta).reset_index(drop=True)
        return self.to_df(ndf[keys + [dest_col_name]])

    def _window_simple_agg(  # noqa: C901
        self,
        agg_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(
            not (func.has_order_by and not func.has_partition_by),
            ValueError("for dask engine, order by requires partition by"),
        )
        assert_or_throw(
            func.has_partition_by,
            ValueError(
                "for dask engine, partition by is required for window aggregation"
            ),
        )
        assert_or_throw(len(args) == 1 and args[0].is_col, ValueError(f"{args}"))
        assert_or_throw(
            not func.unique and not func.dropna,
            NotImplementedError(
                f"window function doesn't support unique {func.unique} "
                + f"and dropna {func.dropna}"
            ),
        )
        col = args[0].value

        ndf = self.to_native(df)
        keys = list(df.keys())
        ndf, sort_keys, asc = self._prepare_sort(ndf, func.window.order_by)
        gp, gp_keys, mi = self._safe_groupby(ndf, func.window.partition_keys)

        def gp_apply(gdf: Any) -> Any:
            if len(sort_keys) > 0:
                gdf = gdf.sort_values(sort_keys, ascending=asc)
            if not func.has_windowframe:
                res: Any = agg_func(gdf[col])
            else:
                wf = func.window.windowframe
                assert_or_throw(
                    isinstance(wf, RollingWindowFrame), NotImplementedError(f"{wf}")
                )
                min_p = 0 if func.name in ["count", "min", "max"] else 1
                # 1.1.0 has this bug
                # https://github.com/pandas-dev/pandas/issues/35557
                # we try to use native functions here to bypass bugs
                if wf.start is None and wf.end == 0:
                    rolling_func: Callable = lambda x: x.expanding(min_periods=min_p)
                elif wf.window is not None and wf.end == 0:  # type: ignore
                    rolling_func: Callable = lambda x: x.rolling(  # type: ignore
                        window=wf.window, min_periods=min_p  # type: ignore
                    )
                else:  # will not work on version pandas 1.1.0  # pragma: no cover
                    rolling_func: Callable = lambda x: x.rolling(  # type: ignore
                        window=_RowsIndexer(wf.start, wf.end), min_periods=min_p
                    )
                res = agg_func(rolling_func(gdf[col]))
            gdf[dest_col_name] = res
            return gdf.reset_index(drop=True)

        meta = [(ndf[x].name, ndf[x].dtype) for x in ndf.columns]
        meta += [(dest_col_name, ndf[col].dtype)]

        ndf = gp.apply(gp_apply, meta=meta).reset_index(drop=True)
        return self.to_df(ndf[keys + [dest_col_name]])

    def _group_agg_with_keys(  # noqa: C901
        self,
        df: DataFrame,
        keys: List[str],
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        ndf = self.to_native(df)
        gp, gp_keys, mi = self._safe_groupby(ndf, keys)
        col_group1: List[pd.Series] = []
        col_group2: List[pd.Series] = []
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                continue
            series, is_agg = self._agg_gp(v[1], gp[v[0]], ndf[v[0]].dtype, mi)
            series = series.rename(k)
            if is_agg:
                col_group1.append(series)
            else:  # different divisions, will need special handling...
                col_group2.append(series)
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                if v[1].unique:
                    series = self._count_unique(ndf, gp_keys, v[0], v[1]).rename(k)
                    col_group2.append(series)
                else:
                    series = self._count_all(gp, v[1]).rename(k)
                    col_group1.append(series)
        res: Any = None
        if len(col_group1) > 0:
            res = col_group1[0].to_frame()
            for s in col_group1[1:]:
                res[s.name] = s
            # TODO: the index of group agg is incorrect in dask
            # otherwise, we can join by index, that can be faster
            res = res.reset_index()
        if len(col_group2) > 0:
            if res is None:
                res = col_group2[0].reset_index()
                n = 1
            else:
                n = 0
            for s in col_group2[n:]:
                tdf = s.reset_index()
                res = res.merge(tdf, "outer", on=gp_keys)
        assert_or_throw(res is not None, ValueError("No aggregation happened"))
        res = res.reset_index()
        res = res[list(agg_map.keys())]
        return self.to_df(res)

    def _group_agg_no_keys(
        self,
        df: DataFrame,
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        ndf = self.to_native(df)
        aggs: Dict[str, Any] = {}
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                continue
            aggs[k] = self._agg_no_gp(v[1], ndf[v[0]])
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                if v[1].unique:
                    aggs[k] = self._count_unique_no_group(ndf, v[0], v[1])
                else:
                    aggs[k] = self._count_all_no_group(ndf, v[1])
        data: Dict[str, Any] = {}
        for k, vv in aggs.items():
            s = vv.compute() if hasattr(vv, "compute") else vv
            if isinstance(s, pandas.Series):
                data[k] = [s.iloc[0]]
            else:
                data[k] = [s]
        return self.to_df(pandas.DataFrame(data))

    def _agg_gp(  # noqa: C901
        self, func: AggFunctionSpec, gp: Any, dtype: Any, index: pandas.MultiIndex
    ) -> Tuple[Any, bool]:
        # TODO: this function has compatibility issue with pandas < 1
        # for example: we want to compute the sum of a same column twice
        # https://github.com/pandas-dev/pandas/issues/27519
        name = func.name.lower()
        if name == "sum":
            if not func.unique:
                return gp.sum(min_count=1), True
            else:
                return (
                    gp.apply(
                        lambda x: x.drop_duplicates().sum(min_count=1),
                        meta=pandas.Series(dtype=dtype, index=index),
                    ),
                    False,
                )
        if name in ["avg", "mean"]:
            if not func.unique:
                return gp.mean(), True
            else:
                return (
                    gp.apply(
                        lambda x: x.drop_duplicates().mean(),
                        meta=pandas.Series(dtype="f8", index=index),
                    ),
                    False,
                )
        if name in ["first", "first_value"]:
            if func.dropna:
                return gp.first(), True
            else:
                # in distributed system, there is no strict order
                # so first or last can be equivalent to choosing any value
                return gp.first(), True
        if name in ["last", "last_value"]:
            if func.dropna:
                return gp.last(), True
            else:
                # in distributed system, there is no strict order
                # so first or last can be equivalent to choosing any value
                return gp.last(), True
        if name == "count":
            if not func.unique and not func.dropna:
                return gp.size(), True
            if func.unique:
                if func.dropna:
                    return gp.nunique(), True
                else:
                    return (
                        gp.apply(
                            lambda x: x.drop_duplicates().size,
                            meta=pandas.Series(dtype="i8", index=index),
                        ),
                        False,
                    )
            if not func.unique and func.dropna:
                return gp.count(), True
        if name == "min":
            if numpy.issubdtype(dtype, numpy.number):
                return gp.min(), True

            def min_(s: Any) -> Any:
                return s.apply(lambda x: x.dropna().min())

            agg = pd.Aggregation("min", min_, min_)
            return gp.agg(agg), True
        if name == "max":
            if numpy.issubdtype(dtype, numpy.number):
                return gp.max(), True

            def max_(s: Any) -> Any:
                return s.apply(lambda x: x.dropna().max())

            agg = pd.Aggregation("max", max_, max_)
            return gp.agg(agg), True
        raise NotImplementedError  # pragma: no cover

    def _count_all(self, df_agg: Any, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and not func.unique,
            ValueError(f"{func} is invalid"),
        )
        return df_agg.size()

    def _count_unique(
        self, df: Any, keys: List[str], cols: str, func: AggFunctionSpec
    ) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and func.unique,
            ValueError(f"{func} can't be applied on {cols}"),
        )
        if cols == "*":
            dkeys = set(df.columns)
        else:
            dkeys = set(keys)
            dkeys.update(cols.split(","))
        return df[list(dkeys)].drop_duplicates().groupby(keys).size()

    def _agg_no_gp(self, func: AggFunctionSpec, x: Any) -> Any:  # noqa: C901
        name = func.name.lower()
        if name == "sum":
            # bug https://github.com/dask/dask/issues/6563
            if not func.unique:
                return x.to_frame().sum(min_count=1).compute()[0]
            else:
                return x.to_frame().drop_duplicates().sum(min_count=1).compute()[0]
        if name in ["avg", "mean"]:
            if not func.unique:
                return x.mean()
            else:
                return x.drop_duplicates().mean()
        if name in ["first", "first_value"]:
            if func.dropna:
                return x.dropna().head(1, compute=False)
            else:
                return x.head(1, compute=False)
        if name in ["last", "last_value"]:
            if func.dropna:
                return x.dropna().tail(1, compute=False)
            else:
                return x.tail(1, compute=False)
        if name == "count":
            if not func.unique and not func.dropna:
                return x.size
            if func.unique and not func.dropna:
                return x.drop_duplicates().size
            if func.unique and func.dropna:
                return x.nunique()
            if not func.unique and func.dropna:
                return x.count()
        if name == "min":
            return x.dropna().min()
        if name == "max":
            return x.dropna().max()
        raise NotImplementedError  # pragma: no cover

    def _count_all_no_group(self, df: Any, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and not func.unique,
            ValueError(f"{func} is invalid"),
        )
        return df.shape[0]

    def _count_unique_no_group(self, df: Any, cols: str, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and func.unique,
            ValueError(f"{func} can't be applied on {cols}"),
        )
        if cols == "*":
            dkeys = set(df.columns)
        else:
            dkeys = set(cols.split(","))
        return df[list(dkeys)].drop_duplicates().shape[0]

    def _safe_groupby(
        self, ndf: Any, keys: List[str]
    ) -> Tuple[Any, List[str], pandas.MultiIndex]:
        nulldf = ndf[keys].isnull()
        gp_keys: List[str] = []
        orig_keys = list(ndf.columns)
        for k in keys:
            ndf[k + "_n"] = nulldf[k]
            ndf[k + "_g"] = ndf[k].fillna(0)
            gp_keys.append(k + "_n")
            gp_keys.append(k + "_g")
        ndf = ndf[orig_keys + gp_keys]
        mi_data = [numpy.ndarray(0, ndf[x].dtype) for x in gp_keys]
        mi = pandas.MultiIndex.from_arrays(mi_data, names=gp_keys)
        return ndf.groupby(gp_keys), gp_keys, mi

    def _safe_pandas_groupby(
        self,
        ndf: pandas.DataFrame,
        keys: List[str],
        keep_extra_keys: bool = False,
        as_index: bool = True,
        sort: bool = True,
    ) -> Tuple[Any, List[str]]:
        if not keep_extra_keys:
            ndf = ndf[ndf.columns]
        nulldf = ndf[keys].isnull()
        gp_keys: List[str] = []
        for k in keys:
            ndf[k + "_n"] = nulldf[k]
            ndf[k + "_g"] = ndf[k].fillna(0)
            gp_keys.append(k + "_n")
            gp_keys.append(k + "_g")
        return ndf.groupby(gp_keys, as_index=as_index, sort=sort), gp_keys

    def _prepare_sort(
        self, ndf: Any, order_by: OrderBySpec
    ) -> Tuple[Any, List[str], List[bool]]:
        if len(order_by.keys) == 0:
            return ndf, [], []
        okeys: List[str] = []
        asc: List[bool] = []
        for oi in order_by:
            nk = oi.name + "_null"
            ndf[nk] = ndf[oi.name].isnull()
            okeys.append(nk)
            asc.append(oi.pd_na_position != "first")
            okeys.append(oi.name)
            asc.append(oi.asc)
        return ndf, okeys, asc



from qpd.workflow import QPDWorkflow, QPDWorkflowContext, WorkflowDataFrame
from qpd_pandas import QPDPandasEngine
import pandas as pd
from pytest import raises


def test_workflow():
    ctx = QPDWorkflowContext(
        QPDPandasEngine(),
        {
            "x": pd.DataFrame([[0, 1], [2, 3]], columns=["a", "b"]),
            "y": pd.DataFrame([[True], [False]], columns=["aa"]),
        },
    )
    wf = QPDWorkflow(ctx)
    df = wf.dfs["x"]
    # invalid to add non workflow dfs
    raises(ValueError, lambda: WorkflowDataFrame(df, ctx.dfs["y"]))
    assert 5 == len(wf._spec.tasks)  # x: 2+1, y: 1+1
    df = df[["b", "a"]]
    assert 6 == len(wf._spec.tasks)
    df = wf.op_to_df(["b", "a"], "filter_df", df, wf.dfs["y"]["aa"])
    assert 10 == len(wf._spec.tasks)
    col1 = df["a"]
    col2 = df["b"]
    col4 = wf.op_to_col("binary_arithmetic_op", col1, col2, "+").rename("p")
    col5 = wf.const_to_col(3, "y").rename("y")  # rename will not take effect
    assert 13 == len(wf._spec.tasks)
    wf.assemble_output(col4, col5)
    wf.assemble_output(col4, col5)  # dedupped
    assert 15 == len(wf._spec.tasks)

    assert [[1, 3]] == wf.run().values.tolist()
